////////////////////////////////////////////////////////////////////////////
//                           **** WAVPACK ****                            //
//                  Hybrid Lossless Wavefile Compressor                   //
//              Copyright (c) 1998 - 2004 Conifer Software.               //
//                          All Rights Reserved.                          //
//      Distributed under the BSD Software License (see license.txt)      //
////////////////////////////////////////////////////////////////////////////

// bits.c

// This module provides utilities to support the BitStream structure which is
// used to read and write all WavPack audio data streams. It also contains a
// wrapper for the stream I/O functions and a set of functions dealing with
// endian-ness, both for enhancing portability. Finally, a debug wrapper for
// the malloc() system is provided.

#include "wavpack.h"

#include <stdlib.h>
#include <process.h>
#include <string.h>
#include <ctype.h>
#include <io.h>
#include <sys\stat.h>

////////////////////////// Bitstream functions ////////////////////////////////

// Open the specified BitStream and associate with the specified buffer.

static void bs_read (Bitstream *bs);

void bs_open_read (Bitstream *bs, uchar *buffer_start, uchar *buffer_end)
{
    bs->error = bs->sr = bs->bc = 0;
    bs->ptr = (bs->buf = buffer_start) - 1;
    bs->end = buffer_end;
    bs->wrap = bs_read;
}

// This function is only called from the getbit() and getbits() macros when
// the BitStream has been exhausted and more data is required. Sinve these
// bistreams no longer access files, this function simple sets an error and
// resets the buffer.

static void bs_read (Bitstream *bs)
{
    bs->ptr = bs->buf - 1;
    bs->error = 1;
}

// This function is called to close the bitstream. It returns the number of
// full bytes actually read as bits.

ulong bs_close_read (Bitstream *bs)
{
    ulong bytes_read;

    if (bs->bc < 8)
	bs->ptr++;

    if ((bs->buf - bs->ptr) & 1)
	bs->ptr++;

    bytes_read = bs->ptr - bs->buf;
    CLEAR (*bs);
    return bytes_read;
}

//////////////////////////// File I/O Wrapper ////////////////////////////////

int DoReadFile (reader *hFile, void *lpBuffer, ulong nNumberOfBytesToRead, ulong *lpNumberOfBytesRead)
{
    ulong bcount;

    *lpNumberOfBytesRead = 0;

    while (nNumberOfBytesToRead) {
	bcount = hFile->read ((uchar *) lpBuffer + *lpNumberOfBytesRead, nNumberOfBytesToRead);

	if (bcount) {
	    *lpNumberOfBytesRead += bcount;
	    nNumberOfBytesToRead -= bcount;
	}
	else
	    break;
    }

    return 1;
}

ulong DoGetFileSize (reader *hFile)
{
    ulong length;

    if (!hFile || (length = (ulong) hFile->get_length()) == -1)
	return 0;

    return length;
}

ulong DoGetFilePosition (reader *hFile)
{
    return (ulong) hFile->get_position ();
}

int DoSetFilePositionAbsolute (reader *hFile, ulong pos)
{
    return !hFile->seek (pos);
}

int DoSetFilePositionRelative (reader *hFile, long pos, int mode)
{
    return !hFile->seek2 (pos, mode);
}

// if ungetc() is not available, a seek of -1 is fine also because we do not
// change the byte read.

int DoUngetc (int c, reader *hFile)
{
    return hFile->seek2 (-1, SEEK_CUR);
}

int DoCloseHandle (reader *hFile)
{
    hFile->reader_release ();
	return 1;
}

/////////////////////// Endian Correction Routines ////////////////////////////

void little_endian_to_native (void *data, char *format)
{
    uchar *cp = (uchar *) data;
    long temp;

    while (*format) {
	switch (*format) {
	    case 'L':
		temp = cp [0] + (cp [1] << 8) + ((long) cp [2] << 16) + ((long) cp [3] << 24);
		* (long *) cp = temp;
		cp += 4;
		break;

	    case 'S':
		temp = cp [0] + (cp [1] << 8);
		* (short *) cp = (short) temp;
		cp += 2;
		break;

	    default:
		if (isdigit (*format))
		    cp += *format - '0';

		break;
	}

	format++;
    }
}

void native_to_little_endian (void *data, char *format)
{
    uchar *cp = (uchar *) data;
    long temp;

    while (*format) {
	switch (*format) {
	    case 'L':
		temp = * (long *) cp;
		*cp++ = (uchar)temp;
		*cp++ = (uchar)(temp >> 8);
		*cp++ = (uchar)(temp >> 16);
		*cp++ = (uchar)(temp >> 24);
		break;

	    case 'S':
		temp = * (short *) cp;
		*cp++ = (uchar)temp;
		*cp++ = (uchar)(temp >> 8);
		break;

	    default:
		if (isdigit (*format))
		    cp += *format - '0';

		break;
	}

	format++;
    }
}

////////////////////////// Debug Wrapper for Malloc ///////////////////////////

#ifdef DEBUG_ALLOC

long malloc_count;

void *malloc_db (ulong size)
{
    if (size) {
	malloc_count++;
	return malloc (size);
    }
    else
	return NULL;
}

void *realloc_db (void *ptr, ulong size)
{
    if (ptr && size)
	return realloc (ptr, size);
    else if (size) {
	malloc_count++;
	return malloc (size);
    }
    else if (ptr) {
	malloc_count--;
	return NULL;
    }

    return NULL;
}

void free_db (void *ptr)
{
    if (ptr) {
	malloc_count--;
	free (ptr);
    }
}

long dump_alloc (void)
{
    return malloc_count;
}

#endif
