////////////////////////////////////////////////////////////////////////////
//			     **** WAVPACK ****				  //
//		    Hybrid Lossless Wavefile Compressor			  //
//		Copyright (c) 1998 - 2004 Conifer Software.		  //
//			    All Rights Reserved.			  //
//      Distributed under the BSD Software License (see license.txt)      //
////////////////////////////////////////////////////////////////////////////

// wputils.c

// This module provides a high-level interface to reading WavPack files.
// WavPack files are opened as standard "C" streams. However, provisions have
// been made to make using another file I/O method easy. The only function
// called directly from here is fopen (); all others are called indirectly
// through a wrapper provided in the bits.c module.

#include "wavpack.h"

#include <stdlib.h>
#include <stdio.h>
#include <io.h>
#include <fcntl.h>
#include <string.h>
#include <math.h>

#ifdef DEBUG_ALLOC
#define malloc malloc_db
#define realloc realloc_db
#define free free_db
void *malloc_db (ulong size);
void *realloc_db (void *ptr, ulong size);
void free_db (void *ptr);
long dump_alloc (void);
#endif

///////////////////////////// local table storage ////////////////////////////

const ulong sample_rates [] = { 6000, 8000, 9600, 11025, 12000, 16000, 22050,
    24000, 32000, 44100, 48000, 64000, 88200, 96000, 192000 };

///////////////////////////// executable code ////////////////////////////////

static ulong read_next_header (reader *infile, WavpackHeader *wphdr);

// This function attempts to open the specified WavPack file for reading. If
// this fails for any reason then an appropriate message is copied to "error"
// and NULL is returned, otherwise a pointer to a WavpackContext structure is
// returned (which is used to call all other functions in this module). A
// filename beginning with "-" is assumed to be stdin. The "flags" argument
// has the following bit mask values to specify details of the open operation:

// OPEN_WVC:  attempt to open/read "correction" file
// OPEN_TAGS:  attempt to read ID3v1 / APEv2 tags (requires seekable file)
// OPEN_WRAPPER:  make audio wrapper available (i.e. RIFF) to caller
// OPEN_2CH_MAX:  open only first stream of multichannel file (usually L/R)
// OPEN_NORMALIZE:  normalize floating point data to +/- 1.0

WavpackContext *WavpackOpenFileInput (reader *infile, const char *infilename, char *error, int flags)
{
    WavpackContext *wpc = (WavpackContext *) malloc (sizeof (WavpackContext));
    WavpackStream *wps;
    uchar first_byte;
    ulong bcount;

    if (!wpc)
	return NULL;

    CLEAR (*wpc);
    wpc->infile = infile;
    wpc->total_samples = (ulong) -1;
    wpc->open_flags = flags;

    wpc->filelen = DoGetFileSize (wpc->infile);

    if (!DoReadFile (wpc->infile, &first_byte, 1, &bcount) || bcount != 1) {
	strcpy (error, "can't read all of WavPack file!");
	return WavpackCloseFile (wpc);
    }

    DoUngetc (first_byte, wpc->infile);

#ifdef VER3
    if (first_byte == 'R')
	return open_file3 (wpc, infilename, error);
#endif

    wpc->streams [0] = wps = (WavpackStream *) malloc (sizeof (WavpackStream));
    wpc->num_streams = 1;
    CLEAR (*wps);

    while (!wps->wphdr.block_samples) {

	wpc->filepos = DoGetFilePosition (wpc->infile);
	bcount = read_next_header (wpc->infile, &wps->wphdr);

	if (bcount == (ulong) -1) {
	    strcpy (error, "not compatible with this version of WavPack file!");
	    return WavpackCloseFile (wpc);
	}

	wpc->filepos += bcount;
	wps->blockbuff = (uchar *) malloc (wps->wphdr.ckSize + 8);
	memcpy (wps->blockbuff, &wps->wphdr, 32);

	if (!DoReadFile (wpc->infile, wps->blockbuff + 32, wps->wphdr.ckSize - 24, &bcount) ||
	    bcount != wps->wphdr.ckSize - 24) {
		strcpy (error, "can't read all of WavPack file!");
		return WavpackCloseFile (wpc);
	}

	if ((wps->wphdr.flags & UNKNOWN_FLAGS) || wps->wphdr.version < 0x402 || wps->wphdr.version > 0x40f) {
	    strcpy (error, "not compatible with this version of WavPack file!");
	    return WavpackCloseFile (wpc);
	}

	if (wps->wphdr.block_samples && wps->wphdr.total_samples != (ulong) -1)
	    wpc->total_samples = wps->wphdr.total_samples;

	if (wps->wphdr.block_samples && (wps->wphdr.flags & HYBRID_FLAG) && (flags & OPEN_WVC)) {
	    string8 in2filename = infilename;

	    in2filename += "c";

	    wpc->in2file = file::g_open (in2filename, reader::MODE_READ);

	    if (wpc->in2file) {
		wpc->file2len = DoGetFileSize (wpc->in2file);
		wpc->wvc_flag = TRUE;
	    }
	}

	if (wpc->wvc_flag) {
	    wpc->file2pos = DoGetFilePosition (wpc->in2file);
	    bcount = read_next_header (wpc->in2file, &wps->wphdr);

	    if (bcount == (ulong) -1) {
		strcpy (error, "problem with correction file");
		return WavpackCloseFile (wpc);
	    }

	    wpc->file2pos += bcount;
	    wps->block2buff = (uchar *) malloc (wps->wphdr.ckSize + 8);
	    memcpy (wps->block2buff, &wps->wphdr, 32);

	    if (!DoReadFile (wpc->in2file, wps->block2buff + 32, wps->wphdr.ckSize - 24, &bcount) ||
		bcount != wps->wphdr.ckSize - 24) {
		    strcpy (error, "can't read all of WavPack file!");
		    return WavpackCloseFile (wpc);
	    }
	}

	if (!unpack_init (wpc)) {
	    strcpy (error, wpc->error_message [0] ? wpc->error_message :
		"not compatible with this version of WavPack file!");

	    return WavpackCloseFile (wpc);
	}
    }

    wpc->config.flags = wps->wphdr.flags & 0xff;
    wpc->config.shift = (wps->wphdr.flags & SHIFT_MASK) >> SHIFT_LSB;

    if (!wpc->config.sample_rate) {
	if (!wps || !wps->wphdr.block_samples || (wps->wphdr.flags & SRATE_MASK) == SRATE_MASK)
	    wpc->config.sample_rate = 44100;
	else
	    wpc->config.sample_rate = sample_rates [(wps->wphdr.flags & SRATE_MASK) >> SRATE_LSB];
    }

    if (!wpc->config.num_channels) {
	wpc->config.num_channels = (wps->wphdr.flags & MONO_FLAG) ? 1 : 2;
	wpc->config.channel_mask = 0x5 - wpc->config.num_channels;
    }

    if ((flags & OPEN_2CH_MAX) && !(wps->wphdr.flags & FINAL_BLOCK))
	wpc->reduced_channels = (wps->wphdr.flags & MONO_FLAG) ? 1 : 2;

    return wpc;
}

// This function obtains general information about an open file and returns
// a mask with the following bit values:

// MODE_WVC:  a .wvc file has been found and will be used for lossless
// MODE_LOSSLESS:  file is lossless (either pure or hybrid)
// MODE_HYBRID:  file is hybrid mode (either lossy or lossless)
// MODE_FLOAT:  audio data is 32-bit ieee floating point
// MODE_VALID_TAG:  file conatins a valid ID3v1 or APEv2 tag

int WavpackGetMode (WavpackContext *wpc)
{
    int mode = 0;

    if (wpc) {
	if (wpc->config.flags & CONFIG_HYBRID_FLAG)
	    mode |= MODE_HYBRID;
	else if (!(wpc->config.flags & CONFIG_LOSSY_MODE))
	    mode |= MODE_LOSSLESS;

	if (wpc->wvc_flag)
	    mode |= (MODE_LOSSLESS | MODE_WVC);

	if (wpc->lossy_blocks)
	    mode &= ~MODE_LOSSLESS;

	if (wpc->config.flags & CONFIG_FLOAT_DATA)
	    mode |= MODE_FLOAT;
    }

    return mode;
}

static void free_streams (WavpackContext *wpc);

// Unpack the specified number of samples from the current file position.
// Note that "samples" here refers to "complete" samples, which would be
// 2 longs for stereo files or even more for multichannel files, so the
// required memory at "buffer" is 4 * samples * num_channels bytes. The
// audio data is returned right-justified in 32-bit longs in the endian
// mode native to the executing processor. So, if the original data was
// 16-bit, then the values returned would be +/-32k. Floating point data
// can also be returned if the source was floating point data (and this
// can be optionally normalized to +/-1.0 by using the appropriate flag
// in the call to WavpackOpenFileInput ()). The actual number of samples
// unpacked is returned, which should be equal to the number requested unless
// the end of fle is encountered or an error occurs.

ulong WavpackUnpackSamples (WavpackContext *wpc, long *buffer, ulong samples)
{
    WavpackStream *wps = wpc->streams [wpc->current_stream = 0];
    ulong bcount, samples_unpacked = 0, samples_to_unpack;
    int num_channels = wpc->config.num_channels;

#ifdef VER3
    if (wpc->stream3)
	return unpack_samples3 (wpc, buffer, samples);
#endif

    while (samples) {
	if (!wps->wphdr.block_samples || !(wps->wphdr.flags & INITIAL_BLOCK) ||
	    wps->sample_index >= wps->wphdr.block_index + wps->wphdr.block_samples) {
		free_streams (wpc);
		wpc->filepos = DoGetFilePosition (wpc->infile);
		bcount = read_next_header (wpc->infile, &wps->wphdr);

		if (bcount == (ulong) -1)
		    break;

		wpc->filepos += bcount;
		wps->blockbuff = (uchar *) malloc (wps->wphdr.ckSize + 8);
		memcpy (wps->blockbuff, &wps->wphdr, 32);

		if (!DoReadFile (wpc->infile, wps->blockbuff + 32, wps->wphdr.ckSize - 24, &bcount) ||
		    bcount != wps->wphdr.ckSize - 24) {
			strcpy (wpc->error_message, "can't read all of WavPack file!");
			break;
		}

		if (wps->wphdr.version < 0x402 || wps->wphdr.version > 0x40f) {
		    strcpy (wpc->error_message, "not compatible with this version of WavPack file!");
		    break;
		}

		if (wps->wphdr.block_samples && wpc->wvc_flag) {
		    wpc->file2pos = DoGetFilePosition (wpc->in2file);
		    bcount = read_next_header (wpc->in2file, &wps->wphdr);

		    if (bcount == (ulong) -1)
			break;

		    wpc->file2pos += bcount;
		    wps->block2buff = (uchar *) malloc (wps->wphdr.ckSize + 8);
		    memcpy (wps->block2buff, &wps->wphdr, 32);

		    if (!DoReadFile (wpc->in2file, wps->block2buff + 32, wps->wphdr.ckSize - 24, &bcount) ||
			bcount != wps->wphdr.ckSize - 24)
			    break;
		}

		if (wps->sample_index == wps->wphdr.block_index && !unpack_init (wpc))
		    break;
	}

	if (!wps->wphdr.block_samples || !(wps->wphdr.flags & INITIAL_BLOCK) ||
	    wps->sample_index >= wps->wphdr.block_index + wps->wphdr.block_samples)
		continue;

	if (wps->sample_index < wps->wphdr.block_index) {
	    samples_to_unpack = wps->wphdr.block_index - wps->sample_index;

	    if (samples_to_unpack > samples)
		samples_to_unpack = samples;

	    wps->sample_index += samples_to_unpack;
	    samples_unpacked += samples_to_unpack;
	    samples -= samples_to_unpack;

	    if (wpc->reduced_channels)
		samples_to_unpack *= wpc->reduced_channels;
	    else
		samples_to_unpack *= num_channels;

	    while (samples_to_unpack--)
		*buffer++ = 0;

	    continue;
	}

	samples_to_unpack = wps->wphdr.block_index + wps->wphdr.block_samples - wps->sample_index;

	if (samples_to_unpack > samples)
	    samples_to_unpack = samples;

	if (!wpc->reduced_channels && !(wps->wphdr.flags & FINAL_BLOCK)) {
	    long *temp_buffer = (long *) malloc (samples_to_unpack * 8), *src, *dst;
	    int offset = 0;
	    ulong samcnt;

	    while (1) {
		if (wpc->current_stream == wpc->num_streams) {
		    wps = wpc->streams [wpc->num_streams++] = (WavpackStream *) malloc (sizeof (WavpackStream));
		    CLEAR (*wps);
		    bcount = read_next_header (wpc->infile, &wps->wphdr);

		    if (bcount == (ulong) -1)
			break;

		    wps->blockbuff = (uchar *) malloc (wps->wphdr.ckSize + 8);
		    memcpy (wps->blockbuff, &wps->wphdr, 32);

		    if (!DoReadFile (wpc->infile, wps->blockbuff + 32, wps->wphdr.ckSize - 24, &bcount) ||
			bcount != wps->wphdr.ckSize - 24) {
			    strcpy (wpc->error_message, "can't read all of WavPack file!");
			    break;
		    }

		    if (wpc->wvc_flag) {
			bcount = read_next_header (wpc->in2file, &wps->wphdr);

			if (bcount == (ulong) -1)
			    break;

			wps->block2buff = (uchar *) malloc (wps->wphdr.ckSize + 8);
			memcpy (wps->block2buff, &wps->wphdr, 32);

			if (!DoReadFile (wpc->in2file, wps->block2buff + 32, wps->wphdr.ckSize - 24, &bcount) ||
			    bcount != wps->wphdr.ckSize - 24)
				break;
		    }

		    if (!unpack_init (wpc))
			break;
		}
		else
		    wps = wpc->streams [wpc->current_stream];

		unpack_samples (wpc, src = temp_buffer, samples_to_unpack);
		samcnt = samples_to_unpack;
		dst = buffer + offset;

		if (wps->wphdr.flags & MONO_FLAG) {
		    while (samcnt--) {
			dst [0] = *src++;
			dst += num_channels;
		    }

		    offset++;
		}
		else {
		    while (samcnt--) {
			dst [0] = *src++;
			dst [1] = *src++;
			dst += num_channels;
		    }

		    offset += 2;
		}
	
		if (wps->wphdr.flags & FINAL_BLOCK)
		    break;
		else
		    wpc->current_stream++;
	    }

	    wps = wpc->streams [wpc->current_stream = 0];
	    free (temp_buffer);
	}
	else
	    unpack_samples (wpc, buffer, samples_to_unpack);

	if (wpc->reduced_channels)
	    buffer += samples_to_unpack * wpc->reduced_channels;
	else
	    buffer += samples_to_unpack * num_channels;

	samples_unpacked += samples_to_unpack;
	samples -= samples_to_unpack;

	if (wps->sample_index == wps->wphdr.block_index + wps->wphdr.block_samples) {
	    if (check_crc_error (wpc) && wps->blockbuff) {
		long rseek = ((WavpackHeader *) wps->blockbuff)->ckSize / 3;

		DoSetFilePositionRelative (wpc->infile, (rseek > 16384) ? -16384 : -rseek, SEEK_CUR);

		if (wpc->wvc_flag && wps->block2buff) {
		    rseek = ((WavpackHeader *) wps->block2buff)->ckSize / 3;
		    DoSetFilePositionRelative (wpc->in2file, (rseek > 16384) ? -16384 : -rseek, SEEK_CUR);
		}

		wpc->crc_errors++;
	    }
	}

	if (wps->sample_index == wpc->total_samples)
	    break;
    }

    return samples_unpacked;
}

#ifdef SEEKING

static ulong find_sample (WavpackContext *wpc, reader *infile, ulong header_pos, ulong sample);

// Seek to the specifed sample index, returning TRUE on success. Note that
// files generated with version 4.0 or newer will seek almost immediately.
// Older files can take quite long if required to seek through unplayed
// portions of the file, but will create a seek map so that reverse seeks
// (or forward seeks to already scanned areas) will be very fast.

int WavpackSeekSample (WavpackContext *wpc, ulong sample)
{
    WavpackStream *wps = wpc->streams [wpc->current_stream = 0];
    ulong bcount, samples_to_skip;
    long *buffer;

#ifdef VER3
    if (wpc->stream3)
	return seek_sample3 (wpc, sample);
#endif

    if (!wps->wphdr.block_samples || !(wps->wphdr.flags & INITIAL_BLOCK) || sample < wps->wphdr.block_index ||
	sample >= wps->wphdr.block_index + wps->wphdr.block_samples) {

	    free_streams (wpc);
	    wpc->filepos = find_sample (wpc, wpc->infile, wpc->filepos, sample);

	    if (wpc->filepos == (ulong) -1)
		return FALSE;

	    if (wpc->wvc_flag) {
		wpc->file2pos = find_sample (wpc, wpc->in2file, 0, sample);

		if (wpc->file2pos == (ulong) -1)
		    return FALSE;
	    }
    }

    if (!wps->blockbuff) {
	DoSetFilePositionAbsolute (wpc->infile, wpc->filepos);
	DoReadFile (wpc->infile, &wps->wphdr, sizeof (WavpackHeader), &bcount);
	little_endian_to_native (&wps->wphdr, WavpackHeaderFormat);
	wps->blockbuff = (uchar *) malloc (wps->wphdr.ckSize + 8);
	memcpy (wps->blockbuff, &wps->wphdr, sizeof (WavpackHeader));

	if (!DoReadFile (wpc->infile, wps->blockbuff + sizeof (WavpackHeader), wps->wphdr.ckSize - 24, &bcount) ||
	    bcount != wps->wphdr.ckSize - 24) {
		free_streams (wpc);
		return FALSE;
	}

	if (wpc->wvc_flag) {
	    DoSetFilePositionAbsolute (wpc->in2file, wpc->file2pos);
	    DoReadFile (wpc->in2file, &wps->wphdr, sizeof (WavpackHeader), &bcount);
	    little_endian_to_native (&wps->wphdr, WavpackHeaderFormat);
	    wps->block2buff = (uchar *) malloc (wps->wphdr.ckSize + 8);
	    memcpy (wps->block2buff, &wps->wphdr, sizeof (WavpackHeader));

	    if (!DoReadFile (wpc->in2file, wps->block2buff + sizeof (WavpackHeader), wps->wphdr.ckSize - 24, &bcount) ||
		bcount != wps->wphdr.ckSize - 24) {
		    free_streams (wpc);
		    return FALSE;
	    }
	}

	if (!unpack_init (wpc)) {
	    free_streams (wpc);
	    return FALSE;
	}
    }

    while (!wpc->reduced_channels && !(wps->wphdr.flags & FINAL_BLOCK)) {
	if (++wpc->current_stream == wpc->num_streams) {
	    wps = wpc->streams [wpc->num_streams++] = (WavpackStream *) malloc (sizeof (WavpackStream));
	    CLEAR (*wps);
	    bcount = read_next_header (wpc->infile, &wps->wphdr);

	    if (bcount == (ulong) -1)
		break;

	    wps->blockbuff = (uchar *) malloc (wps->wphdr.ckSize + 8);
	    memcpy (wps->blockbuff, &wps->wphdr, 32);

	    if (!DoReadFile (wpc->infile, wps->blockbuff + 32, wps->wphdr.ckSize - 24, &bcount) ||
		bcount != wps->wphdr.ckSize - 24) {
		    strcpy (wpc->error_message, "can't read all of WavPack file!");
		    break;
	    }

	    if (wpc->wvc_flag) {
		bcount = read_next_header (wpc->in2file, &wps->wphdr);

		if (bcount == (ulong) -1)
		    break;

		wps->block2buff = (uchar *) malloc (wps->wphdr.ckSize + 8);
		memcpy (wps->block2buff, &wps->wphdr, 32);

		if (!DoReadFile (wpc->in2file, wps->block2buff + 32, wps->wphdr.ckSize - 24, &bcount) ||
		    bcount != wps->wphdr.ckSize - 24)
			break;
	    }

	    if (!unpack_init (wpc))
		break;
	}
	else
	    wps = wpc->streams [wpc->current_stream];
    }

    if (sample < wps->sample_index)
	for (wpc->current_stream = 0; wpc->current_stream < wpc->num_streams; wpc->current_stream++)
	    if (!unpack_init (wpc))
		return FALSE;

    samples_to_skip = sample - wps->sample_index;

    if (samples_to_skip) {
	buffer = (long *) malloc (samples_to_skip * 8);

	for (wpc->current_stream = 0; wpc->current_stream < wpc->num_streams; wpc->current_stream++)
	    unpack_samples (wpc, buffer, samples_to_skip);

	free (buffer);
    }

    wpc->current_stream = 0;
    return TRUE;
}

#endif

// Get total number of samples contained in the WavPack file, or -1 if unknown

ulong WavpackGetNumSamples (WavpackContext *wpc)
{
    return wpc ? wpc->total_samples : (ulong) -1;
}

// Get the current sample index position, or -1 if unknown

ulong WavpackGetSampleIndex (WavpackContext *wpc)
{
    if (wpc) {
#ifdef VER3
	if (wpc->stream3)
	    return get_sample_index3 (wpc);
	else if (wpc->streams [0])
	    return wpc->streams [0]->sample_index;
#else
	if (wpc->streams [0])
	    return wpc->streams [0]->sample_index;
#endif
    }

    return (ulong) -1;
}

// Get the number of errors encountered so far

int WavpackGetNumErrors (WavpackContext *wpc)
{
    return wpc ? wpc->crc_errors : 0;
}

// Calculate the progress through the file as a double from 0.0 (for begin)
// to 1.0 (for done). A return value of -1.0 indicates that the progress is
// unknown.

double WavpackGetProgress (WavpackContext *wpc)
{
    if (wpc && wpc->total_samples != (ulong) -1 && wpc->total_samples != 0)
	return (double) WavpackGetSampleIndex (wpc) / wpc->total_samples;
    else
	return -1.0;
}

// Calculate the ratio of the specified WavPack file size to the size of the
// original audio data as a double greater than 0.0 and (usually) smaller than
// 1.0. A value greater than 1.0 represents "negative" compression and a
// return value of 0.0 indicates that the ratio cannot be determined.

double WavpackGetRatio (WavpackContext *wpc)
{
    if (wpc && wpc->total_samples != (ulong) -1 && wpc->filelen) {
	double output_size = (double) wpc->total_samples * wpc->config.num_channels *
	    ((wpc->config.flags & BYTES_STORED) + 1);
	double input_size = (double) wpc->filelen + wpc->file2len;

	if (output_size >= 1.0 && input_size >= 1.0)
	    return input_size / output_size;
    }

    return 0.0;
}

// Calculate the average bitrate of the WavPack file in bits per second. A
// return of 0.0 indicates that the bitrate cannot be determined.

double WavpackGetAverageBitrate (WavpackContext *wpc)
{
    if (wpc && wpc->total_samples != (ulong) -1 && wpc->filelen) {
	double output_time = (double) wpc->total_samples / wpc->config.sample_rate;
	double input_size = (double) wpc->filelen + wpc->file2len;

	if (output_time >= 1.0 && input_size >= 1.0)
	    return input_size * 8.0 / output_time;
    }

    return 0.0;
}

// Calculate the bitrate of the current WavPack file block in bits per second.
// This can be used for an "instant" bit display and gets updated from about
// 1 to 4 times per second. A return of 0.0 indicates that the bitrate cannot
// be determined.

double WavpackGetInstantBitrate (WavpackContext *wpc)
{
    if (wpc->stream3)
	return WavpackGetAverageBitrate (wpc);

    if (wpc && wpc->streams [0] && wpc->streams [0]->wphdr.block_samples) {
	double output_time = (double) wpc->streams [0]->wphdr.block_samples / wpc->config.sample_rate;
	double input_size = 0;
	int si;

	for (si = 0; si < wpc->num_streams; ++si) {
	    if (wpc->streams [si]->blockbuff)
		input_size += ((WavpackHeader *) wpc->streams [si]->blockbuff)->ckSize;

	    if (wpc->streams [si]->block2buff)
		input_size += ((WavpackHeader *) wpc->streams [si]->block2buff)->ckSize;
	}

	if (output_time > 0.0 && input_size >= 1.0)
	    return input_size * 8.0 / output_time;
    }

    return 0.0;
}

// Close the specified WavPack file and release all resources used by it.
// Returns NULL.

WavpackContext *WavpackCloseFile (WavpackContext *wpc)
{
    free_streams (wpc);

    if (wpc->streams [0])
	free (wpc->streams [0]);

#ifdef VER3
    if (wpc->stream3)
	free_stream3 (wpc);
#endif

    if (wpc->in2file != NULL)		// we must close .wvc file because
	DoCloseHandle (wpc->in2file);	// foobar doesn't know about it...

    free (wpc);

    return NULL;
}

// Returns the sample rate of the specified WavPack file

ulong WavpackGetSampleRate (WavpackContext *wpc)
{
    return wpc ? wpc->config.sample_rate : 44100;
}

// Returns the number of channels of the specified WavPack file. Note that
// this is the actual number of channels contained in the file even if the
// OPEN_2CH_MAX flag was specified when the file was opened.

int WavpackGetNumChannels (WavpackContext *wpc)
{
    return wpc ? wpc->config.num_channels : 2;
}

// Returns the actual number of valid bits per sample contained in the
// original file, which may or may not be a multiple of 8. Floating data
// always has 32 bits, integers may be from 1 to 32 bits each. When this
// value is not a multiple of 8, then the "extra" bits are located in the
// LSBs of the results. That is, values are right justified when unpacked
// into longs, but are left justified in the number of bytes used by the
// original data.

int WavpackGetBitsPerSample (WavpackContext *wpc)
{
    return wpc ? ((wpc->config.flags & CONFIG_BYTES_STORED) + 1) * 8 - wpc->config.shift : 16;
}

// Returns the number of bytes used for each sample (1 to 4) in the original
// file. This is required information for the user of this module because the
// audio data is returned in the LOWER bytes of the long buffer and must be
// left-shifted 8, 16, or 24 bits if normalized longs are required.

int WavpackGetBytesPerSample (WavpackContext *wpc)
{
    return wpc ? (wpc->config.flags & CONFIG_BYTES_STORED) + 1 : 2;
}

// If the OPEN_2CH_MAX flag is specified when opening the file, this function
// will return the actual number of channels decoded from the file (which may
// or may not be less than the actual number of channels, but will always be
// 1 or 2). Normally, this will be the front left and right channels of a
// multichannel file.

int WavpackGetReducedChannels (WavpackContext *wpc)
{
    if (wpc)
	return wpc->reduced_channels ? wpc->reduced_channels : wpc->config.num_channels;
    else
	return 2;
}

// Free all memory allocated for raw WavPack blocks (for all allocated streams)
// and free all additonal streams. This does not free the default stream ([0])
// which is always kept around.

static void free_streams (WavpackContext *wpc)
{
    int si = wpc->num_streams;

    while (si--) {
	if (wpc->streams [si]->blockbuff) {
	    free (wpc->streams [si]->blockbuff);
	    wpc->streams [si]->blockbuff = NULL;
	}

	if (wpc->streams [si]->block2buff) {
	    free (wpc->streams [si]->block2buff);
	    wpc->streams [si]->block2buff = NULL;
	}

	if (si) {
	    wpc->num_streams--;
	    free (wpc->streams [si]);
	    wpc->streams [si] = NULL;
	}
    }

    wpc->current_stream = 0;
}

// Read from current file position until a valid 32-byte WavPack 4.0 header is
// found and read into the specified pointer. The number of bytes skipped is
// returned. If no WavPack header is found within 1 meg, then a -1 is returned
// to indicate the error. No additional bytes are read past the header and it
// is returned in the processor's native endian mode. Seeking is not required.

static ulong read_next_header (reader *infile, WavpackHeader *wphdr)
{
    char buffer [sizeof (*wphdr)], *sp = buffer + sizeof (*wphdr), *ep = sp;
    ulong bcount, bytes_skipped = 0;
    int bleft;

    while (1) {
	if (sp < ep) {
	    bleft = ep - sp;
	    memcpy (buffer, sp, bleft);
	}
	else
	    bleft = 0;

	if (!DoReadFile (infile, buffer + bleft, sizeof (*wphdr) - bleft, &bcount) || bcount != sizeof (*wphdr) - bleft)
	    return -1;

	sp = buffer;

	if (*sp++ == 'w' && *sp == 'v' && *++sp == 'p' && *++sp == 'k' &&
	    !(*++sp & 1) && sp [2] < 16 && !sp [3] && sp [5] == 4 && sp [4] >= 2 && sp [4] <= 0xf) {
		memcpy (wphdr, buffer, sizeof (*wphdr));
		little_endian_to_native (wphdr, WavpackHeaderFormat);
		return bytes_skipped;
	    }

	while (sp < ep && *sp != 'w')
	    sp++;

	if ((bytes_skipped += sp - buffer) > 1024 * 1024)
	    return -1;
    }
}

#ifdef SEEKING

// Find a valid WavPack header, searching either from the current file position
// (or from the specified position if not -1) and store it (endian corrected)
// at the specified pointer. The return value is the exact file position of the
// header, although we may have actually read past it. Because this function
// is used for seeking to a specific audio sample, it only considers blocks
// that contain audio samples for the initial stream to be valid.

#define BUFSIZE 4096

static ulong find_header (reader *infile, ulong filepos, WavpackHeader *wphdr)
{
    char *buffer = (char *) malloc (BUFSIZE), *sp = buffer, *ep = buffer;
    ulong bcount;

    if (filepos != (ulong) -1 && DoSetFilePositionAbsolute (infile, filepos)) {
	free (buffer);
	return -1;
    }

    while (1) {
	int bleft;

	if (sp < ep) {
	    bleft = ep - sp;
	    memcpy (buffer, sp, bleft);
	    ep -= (sp - buffer);
	    sp = buffer;
	}
	else {
	    if (sp > ep)
		if (DoSetFilePositionRelative (infile, sp - ep, SEEK_CUR)) {
		    free (buffer);
		    return -1;
		}

	    sp = ep = buffer;
	    bleft = 0;
	}

	if (!DoReadFile (infile, ep, BUFSIZE - bleft, &bcount))
	    bcount = 0;

	ep += bcount;

	if (ep - sp < 32) {
	    free (buffer);
	    return -1;
	}

	while (sp + 32 <= ep)
	    if (*sp++ == 'w' && *sp == 'v' && *++sp == 'p' && *++sp == 'k' &&
		!(*++sp & 1) && sp [2] < 16 && !sp [3] && sp [5] == 4 && sp [4] >= 2 && sp [4] <= 0xf) {
		    memcpy (wphdr, sp - 4, sizeof (*wphdr));
		    little_endian_to_native (wphdr, WavpackHeaderFormat);

		    if (wphdr->block_samples && (wphdr->flags & INITIAL_BLOCK)) {
			free (buffer);
			return DoGetFilePosition (infile) - (ep - sp + 4);
		    }

		    if (wphdr->ckSize > 1024)
			sp += wphdr->ckSize - 1024;
	    }
    }
}

// Find the WavPack block that contains the specified sample. If "header_pos"
// is zero, then no information is assumed except the total number of samples
// in the file and its size in bytes. If "header_pos" is non-zero then we
// assume that it is the file position of the valid header image contained in
// the first stream and we can limit our search to either the portion above
// or below that point. If a .wvc file is being used, then this must be called
// for that file also.

static ulong find_sample (WavpackContext *wpc, reader *infile, ulong header_pos, ulong sample)
{
    WavpackStream *wps = wpc->streams [wpc->current_stream];
    ulong file_pos1 = 0, file_pos2 = DoGetFileSize (infile);
    ulong sample_pos1 = 0, sample_pos2 = wpc->total_samples;
    double ratio = 0.96;
    int file_skip = 0;

    if (sample >= wpc->total_samples)
	return -1;

    if (header_pos) {
	if (wps->wphdr.block_index > sample) {
	    sample_pos2 = wps->wphdr.block_index;
	    file_pos2 = header_pos;
	}
	else if (wps->wphdr.block_index + wps->wphdr.block_samples <= sample) {
	    sample_pos1 = wps->wphdr.block_index;
	    file_pos1 = header_pos;
	}
	else
	    return header_pos;
    }

    while (1) {
	double bytes_per_sample;
	ulong seek_pos;

	bytes_per_sample = file_pos2 - file_pos1;
	bytes_per_sample /= sample_pos2 - sample_pos1;
	seek_pos = file_pos1 + (file_skip ? 32 : 0);
	seek_pos += (long)(bytes_per_sample * (sample - sample_pos1) * ratio);
	seek_pos = find_header (infile, seek_pos, &wps->wphdr);

	if (seek_pos == (ulong) -1 || seek_pos >= file_pos2) {
	    if (ratio > 0.0) {
		if ((ratio -= 0.24) < 0.0)
		    ratio = 0.0;
	    }
	    else
		return -1;
	}
	else if (wps->wphdr.block_index > sample) {
	    sample_pos2 = wps->wphdr.block_index;
	    file_pos2 = seek_pos;
	}
	else if (wps->wphdr.block_index + wps->wphdr.block_samples <= sample) {

	    if (seek_pos == file_pos1)
		file_skip = 1;
	    else {
		sample_pos1 = wps->wphdr.block_index;
		file_pos1 = seek_pos;
	    }
	}
	else
	    return seek_pos;
    }
}

#endif
